-------------------------------------------------------------------------------
-- File          : blockram.vhd
--
-- Module        : blockram
--
-- Architecture  : Spartan 3E
--
-- Description   : 4KB Dual Port RAM made from block RAM
--
--     Image RAM:  split dual buffered 64 x 96 pixels = 2x8K
--                 Port A is the sampler port - read/write. 
--                 Port B is the scanner port - read only.
--
-- There are two copies of PWM RAM, one for upper nad one for lower.
--     PWM RAM:    64 x 96 pixels = 4K
--                 Only has one port, read/write.
--
-- Revision history:
-- 23Jan2009 DF  Made into tile RAM device
-- 31Mar2009 DF  Changed aspect ratio to 4Kx4 bits * 2 blocks
-- 09Jul2009 DF  Made bigger for dual buffered scan
-- 16Jul2009 DF  Removed lower two bits of scan RAM port
-------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
library unisim;
use unisim.vcomponents.all;

entity blockram is
  
  port (
    clock_i            : in    std_logic;
    reset_i            : in    std_logic;

    -- image RAM sampler port
    samp_write_en      : in    std_logic;
    samp_enable        : in    std_logic;
    samp_addr          : in    std_logic_vector(12 downto 0);
    samp_write_data    : in    std_logic_vector(7 downto 0);
    samp_read_data     : out   std_logic_vector(7 downto 0);

    -- image RAM scanner port
    scan_enable        : in    std_logic;
    scan_addr          : in    std_logic_vector(12 downto 0);
    scan_read_data     : out   std_logic_vector(7 downto 2);

    -- pwm RAM signals (addr is scan_addr)
    pwm_write_en       : in    std_logic;
    pwm_enable         : in    std_logic;
    pwm_addr           : in    std_logic_vector(11 downto 0);
    pwm_write_data     : in    std_logic_vector(7 downto 0);
    pwm_read_data      : out   std_logic_vector(7 downto 0) );

end blockram;

architecture behavioral of blockram is

  constant ground      : std_logic   := '0';
  constant ground_dual : std_logic_vector(1 downto 0) := (others => '0');
  constant ground_nybble : std_logic_vector(3 downto 0) := (others => '0');

  -- prototype for an 8Kx2 dual-port RAM
  component ramb16_s2_s2 
    port (
      wea   : in    std_logic;
      ena   : in    std_logic;
      ssra  : in    std_logic;
      clka  : in    std_logic;
      addra : in    std_logic_vector(12 downto 0);
      dia   : in    std_logic_vector(1 downto 0);
      doa   : out   std_logic_vector(1 downto 0);
      web   : in    std_logic;
      enb   : in    std_logic;
      ssrb  : in    std_logic;
      clkb  : in    std_logic;
      addrb : in    std_logic_vector(12 downto 0);
      dib   : in    std_logic_vector(1 downto 0);
      dob   : out   std_logic_vector(1 downto 0) );
  end component;

  -- prototype for a 8Kx2 single-port RAM
  component ramb16_s2 
    port (
      we   : in    std_logic;
      en   : in    std_logic;
      ssr  : in    std_logic;
      clk  : in    std_logic;
      addr : in    std_logic_vector(12 downto 0);
      di   : in    std_logic_vector(1 downto 0);
      do   : out   std_logic_vector(1 downto 0) );
  end component;

  -- prototype for a 4Kx4 single-port RAM
  component ramb16_s4 
    port (
      we   : in    std_logic;
      en   : in    std_logic;
      ssr  : in    std_logic;
      clk  : in    std_logic;
      addr : in    std_logic_vector(11 downto 0);
      di   : in    std_logic_vector(3 downto 0);
      do   : out   std_logic_vector(3 downto 0) );
  end component;

begin  -- behavioral

-- Port A is for the sampler. Port B is for the readout scanner. 

-- The image RAM is four blocks of 2 bits each. 
-- The lowest two bits are single-ported since never scanned.
    samp_ram_ll: ramb16_s2
      port map (
        we     => samp_write_en,
        en     => samp_enable,
        ssr    => reset_i,
        clk    => clock_i,
        addr   => samp_addr,
        di     => samp_write_data(1 downto 0),
        do     => samp_read_data(1 downto 0) );

    samp_ram_ml: ramb16_s2_s2
      port map (
        wea    => samp_write_en,
        ena    => samp_enable,
        ssra   => reset_i,
        clka   => clock_i,
        addra  => samp_addr,
        dia    => samp_write_data(3 downto 2),
        doa    => samp_read_data(3 downto 2),
        web    => ground,
        enb    => scan_enable,
        ssrb   => reset_i,
        clkb   => clock_i,
        addrb  => scan_addr,
        dib    => ground_dual,
        dob    => scan_read_data(3 downto 2) );

    samp_ram_mh: ramb16_s2_s2
      port map (
        wea    => samp_write_en,
        ena    => samp_enable,
        ssra   => reset_i,
        clka   => clock_i,
        addra  => samp_addr,
        dia    => samp_write_data(5 downto 4),
        doa    => samp_read_data(5 downto 4),
        web    => ground,
        enb    => scan_enable,
        ssrb   => reset_i,
        clkb   => clock_i,
        addrb  => scan_addr,
        dib    => ground_dual,
        dob    => scan_read_data(5 downto 4) );

    samp_ram_hh: ramb16_s2_s2
      port map (
        wea    => samp_write_en,
        ena    => samp_enable,
        ssra   => reset_i,
        clka   => clock_i,
        addra  => samp_addr,
        dia    => samp_write_data(7 downto 6),
        doa    => samp_read_data(7 downto 6),
        web    => ground,
        enb    => scan_enable,
        ssrb   => reset_i,
        clkb   => clock_i,
        addrb  => scan_addr,
        dib    => ground_dual,
        dob    => scan_read_data(7 downto 6) );

-- The pwm RAM is two blocks of 4 bits each. 
    pwm_ram_l: ramb16_s4
      port map (
        we    => pwm_write_en,
        en    => pwm_enable,
        ssr   => reset_i,
        clk   => clock_i,
        addr  => pwm_addr,
        di    => pwm_write_data(3 downto 0),
        do    => pwm_read_data(3 downto 0) );

    pwm_ram_h: ramb16_s4
      port map (
        we    => pwm_write_en,
        en    => pwm_enable,
        ssr   => reset_i,
        clk   => clock_i,
        addr  => pwm_addr,
        di    => pwm_write_data(7 downto 4),
        do    => pwm_read_data(7 downto 4) );

end behavioral;

