--------------------------------------------
--
-- digit.vhd
--
-- Description: Top level of LED TV digitizer chip
--
-- Revision history:
-- 01Apr2009 DF  Made into digit from tile
-- 09Jul2009 DF  Making fancier for big board controller
-- 11Jul2009 DF  Modifying sampler to take hsync input
-- 15Jul2009 DF  Undoing above change
-- 16Jul2009 DF  Removing adc_data_i(0), scan RAM 1:0
-- 01Aug2009 DF  Adding logic analyzer port to test it
-- 09Aug2009 DF  Femoving logic analyzer port for functional version
-- 14Aug2009 DF  Adding clock DCM module to improve clocking
--------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity digit is

  port (
    -- universal inputs
    clock_i            : in    std_logic;
    reset_i            : in    std_logic;

    -- inputs to sampler
    adc_data_i         : in    std_logic_vector(7 downto 1);
    vclk_i             : in    std_logic;
    vsync_i            : in    std_logic;
    hblnk_i            : in    std_logic;
    left_i             : in    std_logic;

    -- interface to upper LED array board
    upper_enab_i       : in    std_logic;
    upper_data_o       : out   std_logic_vector(7 downto 0);
    upper_clk_o        : out   std_logic;
    upper_load_o       : out   std_logic;
    upper_colres_o     : out   std_logic;
    upper_colclk_o     : out   std_logic;

    -- interface to lower LED array board and logic analyzer
    lower_enab_i       : in    std_logic;
    lower_data_o       : out   std_logic_vector(7 downto 0);
    lower_clk_o        : out   std_logic;
    lower_load_o       : out   std_logic;
    lower_colres_o     : out   std_logic;
    lower_colclk_o     : out   std_logic;
    
    -- logic analyzer upper bits are on extra pins
 --   la_clk             : out   std_logic;
    la_d               : out   std_logic_vector(15 downto 12) );

end entity digit;

architecture behavioral of digit is

  signal reset           : std_logic;  -- internal reset signal
  signal good_clk        : std_logic;  -- DCM clock output to stuff
  signal clock_buf       : std_logic;  -- buffered input clock


  component clocker
    port ( CLKIN_IN        : in    std_logic; 
           RST_IN          : in    std_logic; 
           CLKIN_IBUFG_OUT : out   std_logic; 
           CLK0_OUT        : out   std_logic; 
           LOCKED_OUT      : out   std_logic);
  end component;

  component sampler
  port (
    clock_i            : in    std_logic;
    reset_i            : in    std_logic;
    la_d               : out   std_logic_vector(15 downto 12);

    -- Analog Data interface
    adc_data_i         : in    std_logic_vector(7 downto 1);
    vclk_i             : in    std_logic;
    vsync_i            : in    std_logic;
    hblnk_i            : in    std_logic;
    left_i             : in    std_logic;

    disp_field         : out   std_logic;

    -- blockRAM interface - left ports
    samp_write_en      : out   std_logic;
    samp_en_upper      : out   std_logic;
    samp_en_lower      : out   std_logic;
    samp_addr          : out   std_logic_vector(12 downto 0);
    samp_write_data    : out   std_logic_vector(7 downto 0);
    samp_read_data_upper : in  std_logic_vector(7 downto 0);
    samp_read_data_lower : in  std_logic_vector(7 downto 0) );

    end component;

  component scanner
  port (
    clock_i            : in    std_logic;
    reset_i            : in    std_logic;

    -- interface to LED arrays
    strip_enable_i     : in    std_logic;
    strip_data_o       : out   std_logic_vector(7 downto 0);
    strip_clk_o        : out   std_logic;
    strip_load_o       : out   std_logic;
    strip_colres_o     : out   std_logic;
    strip_colclk_o     : out   std_logic;

    disp_field         : in    std_logic;

    -- interface to pixel RAM
    scan_addr          : out   std_logic_vector(12 downto 0);
    scan_enable        : out   std_logic;
    scan_read_data     : in    std_logic_vector(7 downto 2);

    -- interface to pwm RAM
    pwm_addr           : out   std_logic_vector(11 downto 0);
    pwm_enable         : out   std_logic;
    pwm_write_en       : out   std_logic;
    pwm_write_data     : out   std_logic_vector(7 downto 0);
    pwm_read_data      : in    std_logic_vector(7 downto 0) );

  end component;

  component blockram
  port (
    clock_i            : in    std_logic;
    reset_i            : in    std_logic;

    -- image RAM sampler port
    samp_write_en      : in    std_logic;
    samp_enable        : in    std_logic;
    samp_addr          : in    std_logic_vector(12 downto 0);
    samp_write_data    : in    std_logic_vector(7 downto 0);
    samp_read_data     : out   std_logic_vector(7 downto 0);

    -- image RAM scanner port
    scan_enable        : in    std_logic;
    scan_addr          : in    std_logic_vector(12 downto 0);
    scan_read_data     : out   std_logic_vector(7 downto 2);

    -- pwm RAM signals (addr is scan_addr)
    pwm_write_en       : in    std_logic;
    pwm_enable         : in    std_logic;
    pwm_addr           : in    std_logic_vector(11 downto 0);
    pwm_write_data     : in    std_logic_vector(7 downto 0);
    pwm_read_data      : out   std_logic_vector(7 downto 0) );

  end component;

  -- misc internal signals
  signal disp_field         : std_logic;
  
  -- RAM sampler port
  signal samp_write_en      : std_logic;
  signal samp_en_upper      : std_logic;
  signal samp_en_lower      : std_logic;
  signal samp_addr          : std_logic_vector(12 downto 0);
  signal samp_write_data    : std_logic_vector(7 downto 0);
  signal samp_read_data_upper : std_logic_vector(7 downto 0);
  signal samp_read_data_lower : std_logic_vector(7 downto 0);

  -- RAM upper scanner port
  signal uscan_addr         : std_logic_vector(12 downto 0);
  signal uscan_enable       : std_logic;
  signal uscan_read_data    : std_logic_vector(7 downto 2);
  
  -- RAM lower scanner port
  signal lscan_addr         : std_logic_vector(12 downto 0);
  signal lscan_enable       : std_logic;
  signal lscan_read_data    : std_logic_vector(7 downto 2);
  
  -- RAM upper pwm port
  signal upwm_addr          : std_logic_vector(11 downto 0);
  signal upwm_write_en      : std_logic;
  signal upwm_enable        : std_logic;
  signal upwm_write_data    : std_logic_vector(7 downto 0);
  signal upwm_read_data     : std_logic_vector(7 downto 0);

  -- RAM lower pwm port
  signal lpwm_addr          : std_logic_vector(11 downto 0);
  signal lpwm_write_en      : std_logic;
  signal lpwm_enable        : std_logic;
  signal lpwm_write_data    : std_logic_vector(7 downto 0);
  signal lpwm_read_data     : std_logic_vector(7 downto 0);

begin

  CLOCK_COMPONENT: clocker
    port map ( CLKIN_IN    => clock_i, 
           RST_IN          => reset, 
           CLKIN_IBUFG_OUT => clock_buf, 
           CLK0_OUT        => good_clk );

  SAMPLER_COMPONENT: sampler
    port map (
      clock_i              => good_clk,
      reset_i              => reset,
      la_d                 => la_d,

      -- Analog Data interface
      adc_data_i           => adc_data_i,
      vclk_i               => vclk_i,
      vsync_i              => vsync_i,
      hblnk_i              => hblnk_i,
      left_i               => left_i,
      disp_field           => disp_field,

      -- blockRAM interface - left ports
      samp_write_en        => samp_write_en,
      samp_en_upper        => samp_en_upper,
      samp_en_lower        => samp_en_lower,
      samp_addr            => samp_addr,
      samp_write_data      => samp_write_data,
      samp_read_data_upper => samp_read_data_upper,
      samp_read_data_lower => samp_read_data_lower );

  UPPER_SCANNER_COMPONENT: scanner
    port map (
      clock_i              => good_clk,
      reset_i              => reset,

      -- interface to upper LED array
      strip_enable_i       => upper_enab_i,
      strip_data_o         => upper_data_o,
      strip_clk_o          => upper_clk_o,
      strip_load_o         => upper_load_o,
      strip_colres_o       => upper_colres_o,
      strip_colclk_o       => upper_colclk_o,

      disp_field           => disp_field,

      -- interface to image RAM
      scan_addr            => uscan_addr,
      scan_enable          => uscan_enable,
      scan_read_data       => uscan_read_data,

      -- pwm RAM signals (addr is scan_addr)
      pwm_addr             => upwm_addr,
      pwm_write_en         => upwm_write_en,
      pwm_enable           => upwm_enable,
      pwm_write_data       => upwm_write_data,
      pwm_read_data        => upwm_read_data );

  UPPER_BLOCKRAM_COMPONENT: blockram
    port map (
      clock_i              => good_clk,
      reset_i              => reset,

      -- left ports
      samp_write_en        => samp_write_en,
      samp_enable          => samp_en_upper,
      samp_addr            => samp_addr,
      samp_write_data      => samp_write_data,
      samp_read_data       => samp_read_data_upper,

      -- right ports
      scan_addr            => uscan_addr,
      scan_enable          => uscan_enable,
      scan_read_data       => uscan_read_data,

      -- pwm RAM signals (addr is scan_addr)
      pwm_write_en         => upwm_write_en,
      pwm_enable           => upwm_enable,
      pwm_addr             => upwm_addr,
      pwm_write_data       => upwm_write_data,
      pwm_read_data        => upwm_read_data );

  LOWER_SCANNER_COMPONENT: scanner
    port map (
      clock_i              => good_clk,
      reset_i              => reset,

      -- interface to lower LED array
      strip_enable_i       => lower_enab_i,
      strip_data_o         => lower_data_o,
      strip_clk_o          => lower_clk_o,
      strip_load_o         => lower_load_o,
      strip_colres_o       => lower_colres_o,
      strip_colclk_o       => lower_colclk_o,

      disp_field           => disp_field,

      -- interface to image RAM
      scan_addr            => lscan_addr,
      scan_enable          => lscan_enable,
      scan_read_data       => lscan_read_data,

      -- pwm RAM signals (addr is scan_addr)
      pwm_addr             => lpwm_addr,
      pwm_write_en         => lpwm_write_en,
      pwm_enable           => lpwm_enable,
      pwm_write_data       => lpwm_write_data,
      pwm_read_data        => lpwm_read_data );

  LOWER_BLOCKRAM_COMPONENT: blockram
    port map (
      clock_i              => good_clk,
      reset_i              => reset,

      -- left ports
      samp_write_en        => samp_write_en,
      samp_enable          => samp_en_lower,
      samp_addr            => samp_addr,
      samp_write_data      => samp_write_data,
      samp_read_data       => samp_read_data_lower,

      -- right ports
      scan_addr            => lscan_addr,
      scan_enable          => lscan_enable,
      scan_read_data       => lscan_read_data,

      -- pwm RAM signals (addr is scan_addr)
      pwm_write_en         => lpwm_write_en,
      pwm_enable           => lpwm_enable,
      pwm_addr             => lpwm_addr,
      pwm_write_data       => lpwm_write_data,
      pwm_read_data        => lpwm_read_data );

-- *** TEST *** This dummies up the lower RAM scan port. 
-- Replace it with the stuff in lower_logic.txt !!!
      -- right ports
 --     scan_addr            => lscan_addr,
 --     scan_enable          => lscan_enable,
--      scan_read_data       => lscan_read_data,

-- *** end of stuff ***

  -- Define misc connections
  reset                     <= not(reset_i);
--  la_clk                    <= clock_i;

end behavioral;

