-------------------------------------------------------------------------------
-- file: sampler.vhd
--
-- Module: sampler
--
-- Revision history: 
--  23Jan2009 DF  Started with int, changing over to sampler   
--  26Jan2009 DF  Compiles successfully, added daisy chain signals 
--  31Mar2009 DF  Converting to digitizer board FPGA 
--  07Jul2009 DF  Changing to big TV panel controller
--  09Jul2009 DF  Cleanup of new code
--  11Jul2009 DF  converted to 3 state machines
--  15Jul2009 DF  Removed hsync_i, hor_state
--  16Jul2009 DF  Removed adc_data_i(0), made vblank constant
--  02Aug2009 DF  Adding logic analyzer to test sampler
--  09aug2009 DF  Removing logic analyzer for functional lower panel
--  14Aug2009 DF  Adding logic analyzer 12-15 for clock debug
-- 
-- Description: Samples ADC data for one half of screen. 
-- The half-display size is two TV boards, 
-- each of which is 64 pixels wide by 48 pixels tall.
-- The total pixel count is 6K. 
-- Storage is allocated thusly:
-- Two fields worth of image data, for 16K bytes.
-- One field worth of PWM counters, for 8K bytes.
--
-- The NTSC field is downresolved at a rate of 2.5:1. 
-- This is done by adding rows to each other in the image RAM as needed.
-- 
-- The downsampling is done by adding the video data to one or both of 
-- the even and odd image rows. The sequence repeats every 5 video lines.
--
--  when "000"  write 0    + adc/2 to even,  odd + 0     to odd
--  when "001"  write even + adc/2 to even,  odd + 0     to odd
--  when "010"  write even + adc/4 to even,  0   + adc/4 to odd
--  when "011"  write even + 0     to even,  odd + adc/2 to odd
--  when "100"  write even + 0     to even,  odd + adc/2 to odd
-- 
-- State machine descriptions:
--
-- There are two counters implemented as state machines. 
-- These do the horizontal and vertical addressing. 
-- A third state machine controls the data capturing. 
--
-- There are odd and even field halves of the image RAMs to allow one 
-- to be filled while the other is displayed by scanner. 
--
-- There are two image RAMs, upper and lower. This allows the upper and
-- lower display boards to suck from the image RAM simultaneously.
--
--  adc_data_i      is 8 bits of data from ADC
--  vclk_i          is the ADC clock at ~2 MHz
--  vsync_i         is incoming video vertical sync pulse, active low
--  hblnk_i         is true during left end of line
--  left_i          is true if we're the left half of TV
--  samp_write_en   allows writing data to RAM when 1
--  
-- The image RAM is allocated thusly:
--
-- addr bits     contents
--   5-0       pixel across screen (64 pixels)
--     6       image row bit 0 is not in a counter (both done at once)
--  11-7       imrow bits 4-0, counts 0 to 22 by 2
--    12       field for double buffering
--  
-- Some VHDL tricks:
-- 
-- To write a decimal integer value to a vector:
-- 
-- vector <= std_logic_vector(to_unsigned(47,vector'length));
--
-- ---------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
--use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

entity sampler is
  
  port (
    clock_i            : in    std_logic;
    reset_i            : in    std_logic;
    la_d               : out   std_logic_vector(15 downto 12);

    -- Analog Data interface
    adc_data_i         : in    std_logic_vector(7 downto 1);
    vclk_i             : in    std_logic;
    vsync_i            : in    std_logic;
    hblnk_i            : in    std_logic;
    left_i             : in    std_logic;

    disp_field         : out   std_logic;

    -- blockRAM interface - left ports
    samp_write_en      : out   std_logic;
    samp_en_upper      : out   std_logic;
    samp_en_lower      : out   std_logic;
    samp_addr          : out   std_logic_vector(12 downto 0);
    samp_write_data    : out   std_logic_vector(7 downto 0);
    samp_read_data_upper : in  std_logic_vector(7 downto 0);
    samp_read_data_lower : in  std_logic_vector(7 downto 0) );

end sampler;


architecture behavioral of sampler is
  
  -- Vertical state machine counts video lines via hsync. 
  -- It generates line, imrow and subline. 
  signal line                   : std_logic_vector(8 downto 0);
  signal imrow, next_imrow      : std_logic_vector(4 downto 0);
  signal subline, next_subline  : std_logic_vector(2 downto 0);
  
  constant blanks               : std_logic_vector(8 downto 0) := "000010000";

  type vert_state_type is 
       ( vert_blank,           -- in vert blanking interval
         vert_upper,           -- in upper half of active lines
         vert_lower);          -- in lower half of active lines
 
  signal vert_state, next_vert_state : vert_state_type;

  -- Data state machine writes new image data or merges
  -- the incoming image data with the contents of image RAM. 
  -- There are two RMW cycles per pixel, since the 5:2 deresolving
  -- requires the ability to modify both lines of the image data.
  type data_state_type is 
       ( data_vclk,            -- waiting for vclk to go low
         data_read_even,       -- reading even row data
         data_write_even,      -- writing even row data
         data_read_odd,        -- reading odd row data
         data_write_odd);      -- writing odd row data
 
  signal data_state, next_data_state : data_state_type;

  signal image_data, new_data   : std_logic_vector(7 downto 0);
  signal field                  : std_logic;
  signal adc_data               : std_logic_vector(7 downto 1);
  signal pixel                  : std_logic_vector(7 downto 0);
  signal vclk1, vclk2           : std_logic;
  signal vclk,  next_vclk       : std_logic;
  signal vclk_rising            : std_logic;

begin 

  -- This makes the video field toggle.  
  FIELD_SYNC_PROCESS: process (vsync_i, reset_i)
  begin
    if (reset_i = '1') then
      field            <= '0';
    elsif (vsync_i'event and vsync_i = '1') then
      field            <= not field;
    end if;
  end process FIELD_SYNC_PROCESS;

  -- This counter makes video line numbers.  
  -- It's not synchronized to clock_i because its signals
  -- only change when hblnk is true, so their transitions are ignored.
  VERT_SYNC_PROCESS: process (hblnk_i, vsync_i)
  begin
    if (vsync_i = '0') then
      line             <= (others=>'0');
      imrow            <= (others=>'0');
      subline          <= (others=>'0');
      vert_state       <= vert_blank; 
    elsif (hblnk_i'event and hblnk_i = '1') then
      line             <= line + "000000001";
      imrow            <= next_imrow;
      subline          <= next_subline;
      vert_state       <= next_vert_state; 
    end if;
  end process VERT_SYNC_PROCESS;

  VERT_COMB_PROCESS: process (vert_state, line, subline, imrow)
  begin
    -- Line counter incremented by hsync, reset by vsync
    -- counts up vblanking interval, upper half, lower half
    -- vert_state. 
    next_subline   <= subline + "001";
    next_imrow     <= imrow;
    next_vert_state    <= vert_state; 

    case vert_state is
      ---------------------------------------------------------------------
      when vert_blank =>       -- in vert blanking interval
        if (line = blanks) then
          next_vert_state  <= vert_upper; 
          next_subline   <= "000";
          next_imrow     <= "00000";
        end if;
      ---------------------------------------------------------------------
      when vert_upper =>       -- in upper half of active lines
        next_subline   <= subline + "001";
        if (subline = "100") then
          next_subline   <= "000";
          next_imrow     <= imrow + "00001";
        end if;
        if (line = blanks+120) then
          next_vert_state  <= vert_lower; 
          next_subline   <= "000";
          next_imrow     <= "00000";
        end if; 
      ---------------------------------------------------------------------
      when vert_lower =>       -- in lower half of active lines
        next_subline   <= subline + "001";
        if (subline = "100") then
          next_subline   <= "000";
          next_imrow     <= imrow + "00001";
        end if;
        if (line = blanks+240) then
          next_vert_state  <= vert_blank; 
        end if; 
      ---------------------------------------------------------------------
      when others =>
        next_vert_state    <= vert_blank;  -- fail safe
      -----------------------------------------------------------------------
    end case;

  -- This works. All counters go as desired. 
  -- la_d(8 downto 0)            <= line;
  -- la_d(13 downto 9)            <= imrow;
  -- la_d(15 downto 14)            <= subline(1 downto 0);

  end process VERT_COMB_PROCESS;

  -- This counter makes horizontal pixel numbers. 
  -- It's not synchronized to clock_i because its signals
  -- only change when vclk_i is rising, so their transitions are ignored.
  HOR_SYNC_PROCESS: process (vclk_i, hblnk_i)
  begin
    if (hblnk_i = '1') then
      pixel             <= "11111111";  -- not quite pixel 0
    elsif (vclk_i'event and vclk_i = '1') then
      pixel             <= pixel + "00000001";
    end if;
  end process HOR_SYNC_PROCESS;

  -- This state machine writes data into the RAM.

  DATA_SYNC_PROCESS: process (clock_i, reset_i)
  begin
    if (reset_i = '1') then
      data_state       <= data_vclk;
    elsif (clock_i'event and clock_i = '1') then

      -- update various registers
      data_state       <= next_data_state;
      adc_data         <= adc_data_i;

      -- double resync and deglitch the video clock input
      -- because we have to use our fast clock on image data
      vclk1            <= vclk_i;
      vclk2            <= vclk1;
      vclk             <= next_vclk;
    end if;
  end process DATA_SYNC_PROCESS;

  DATA_COMB_PROCESS: process (data_state, vert_state, 
       adc_data, imrow, field, image_data, new_data, 
       left_i, vclk_i, vclk, vclk1, vclk2, pixel, line, subline, 
       samp_read_data_upper, samp_read_data_lower)
  begin
    samp_addr(12 downto 7) <= field & imrow; 
    samp_addr(6)           <= '0'; 
    samp_addr(5 downto 0)  <= pixel(5 downto 0); 
    samp_en_upper        <= '0';
    samp_en_lower        <= '0';
    samp_write_data      <= image_data + new_data;  -- to sum input lines together
    samp_write_en        <= '0';
    disp_field           <= not(field);
    image_data           <= (others=>'0');
    new_data             <= (others=>'0');

    next_data_state      <= data_state;

    -- resync video clock with glitch remover
    next_vclk            <= ((not vclk  and vclk1  and vclk2)  
                           or (vclk  and (vclk1  or vclk2)));
    -- falling edge detector
    vclk_rising          <= not vclk  and vclk1  and vclk2;

    case data_state is
      ---------------------------------------------------------------------
      when data_vclk => -- waiting for vclk to go high in correct half of display
        if (vclk_rising = '1' and vert_state /= vert_blank
           and pixel(6) = left_i and pixel(7) = '0') then
          next_data_state <= data_read_even;
        end if;
      ---------------------------------------------------------------------
      when data_read_even =>         -- reading even row data
        -- enable the correct RAM for this pixel
        if (vert_state = vert_upper) then 
          samp_en_upper  <= '1';
        else
          samp_en_lower  <= '1';
        end if;
        samp_addr(6)     <= '0'; 
        next_data_state  <= data_write_even;
      ---------------------------------------------------------------------
      when data_write_even =>        -- writing even row data
        if (subline /= "000") then
          if (vert_state = vert_upper) then 
            image_data   <= samp_read_data_upper;
          else
            image_data   <= samp_read_data_lower;
          end if;
        end if;

        if (subline = "000" or subline = "001") then
          new_data       <= '0' & adc_data(7 downto 1);
        elsif (subline = "010") then
          new_data       <= "00" & adc_data(7 downto 2);
        end if;
        -- enable the correct RAM for this pixel
        if (vert_state = vert_upper) then 
          samp_en_upper  <= '1';
        else
          samp_en_lower  <= '1';
        end if;
        samp_write_en    <= '1';
        samp_addr(6)     <= '0'; 

        next_data_state  <= data_read_odd;
      ---------------------------------------------------------------------
      when data_read_odd =>          -- reading odd row data
        -- enable the correct RAM for this pixel
        if (vert_state = vert_upper) then 
          samp_en_upper  <= '1';
        else
          samp_en_lower  <= '1';
        end if;
        samp_addr(6)     <= '1'; 
        next_data_state  <= data_write_odd;
      ---------------------------------------------------------------------
      when data_write_odd =>         -- writing odd row data
        if (subline /= "010") then
          if (vert_state = vert_upper) then 
            image_data   <= samp_read_data_upper;
          else
            image_data   <= samp_read_data_lower;
          end if;
        end if;

        if (subline = "011" or subline = "100") then
          new_data       <= '0' & adc_data(7 downto 1);
        elsif (subline = "010") then
          new_data       <= "00" & adc_data(7 downto 2);
        end if;
        -- enable the correct RAM for this pixel
        if (vert_state = vert_upper) then 
          samp_en_upper  <= '1';
        else
          samp_en_lower  <= '1';
        end if;
        samp_write_en    <= '1';
        samp_addr(6)     <= '1'; 
        next_data_state  <= data_vclk;
      ---------------------------------------------------------------------
      when others =>
        next_data_state  <= data_vclk;  -- fail safe
      -----------------------------------------------------------------------
    end case;


 -- la_d(7 downto 0)            <= pixel;
 -- la_d(14 downto 8)          <= samp_read_data_upper(6 downto 0);

  if (data_state = data_write_even) then
    la_d(15)                   <= '1';
  else 
    la_d(15)                   <= '0';
  end if;  
  
  if (data_state = data_read_odd) then
    la_d(14)                   <= '1';
  else 
    la_d(14)                   <= '0';
  end if;  
  
  if (data_state = data_write_odd) then
    la_d(13)                   <= '1';
  else 
    la_d(13)                   <= '0';
  end if;  
  
  if (data_state = data_vclk) then
    la_d(12)                   <= '1';
  else 
    la_d(12)                   <= '0';
  end if;  

  end process DATA_COMB_PROCESS;

end behavioral;

