-------------------------------------------------------------------------------
-- file:   scanner.vhd
--
-- module:  scanner
--
-- Revision history:
-- 23Jan2009 DF  Was read, now scanner
-- 31Mar2009 DF  Converting to make six serial PWM data streams for 12 tile4
-- 07Jul2009 DF  Changing again to 64 wide by 96 tall dual output scheme
-- 14Aug2009 DF  Adding register to data out to improve setup/hold to 595s
-- 15Aug2009 DF  Getting load pulse just ever so
--
-- Description:   Scans pixels from RAM into LEDs using PWM for brightness control
--
--  ****
-- There are two instances of this module, one for upper and the other for lower. 
--  ****
--
-- There are 8 LED array strips on each TV board. 
-- Each strip is 48 pixels tall x 8 pixels wide. 
-- All 8 strips are refreshed in rapid succession. 
-- The master clock is 24 MHz and has two cycles per strip, a read and a write. 
-- The memory array is assigned to the LED matrices in the following order:
-- 
--   led col  0 is the left   end, first after reset. 
--   led col  7 is the right  end, last  after reset.
--   led row  0 is the top    row, last  to leave the serial bus. 
--   led row 47 is the bottom row, first to leave the serial bus. 
--
-- RAM address bit asssignment is as follows:
--   ram_addr(12)    is disp_field   even field if capturing odd field, etc.
--   ram_addr(11..6) is row(5..0)    which of 48 rows in a strip, reverse scan
--   ram_addr(5..3)  is strip(2..0)  which of six strips of tiles
--   ram_addr(2..0)  is col(2..0)    which LED column in a strip
--
-- This module reads port B of the pixel RAM into the PWM RAM, then writes the PWM
-- state for each pixel to the LED strips in serial form. 
-- It runs at 16x the output clock rate - 8 strips * 2 phases.
-- The fastest changing variable is strip, then row, then col, then PWM count. 
--
-- The pwm reload state is when PWM count == row number. 
-- The reason is to stagger the PWM resets to equalize power consumption over time.
-- In pwm reload state, the scan data are copied into the pwm RAM. 
-- The output during this phase is true if scan data was non-zero. 
-- For all other pwm counts, the output data is the MSB of decremented pwm data. 
--
-- The phases are:
--   0  read RAM data
--   1  write back PWM data - 1, inc RAM address
-- 
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_unsigned.all;

entity scanner is

  port (
    clock_i            : in    std_logic;
    reset_i            : in    std_logic;

    -- interface to LED arrays
    strip_enable_i     : in    std_logic;
    strip_data_o       : out   std_logic_vector(7 downto 0);
    strip_clk_o        : out   std_logic;
    strip_load_o       : out   std_logic;
    strip_colres_o     : out   std_logic;
    strip_colclk_o     : out   std_logic;

    disp_field         : in    std_logic;

    -- interface to pixel RAM
    scan_addr          : out   std_logic_vector(12 downto 0);
    scan_enable        : out   std_logic;
    scan_read_data     : in    std_logic_vector(7 downto 2);

    -- interface to pwm RAM
    pwm_addr           : out   std_logic_vector(11 downto 0);
    pwm_enable         : out   std_logic;
    pwm_write_en       : out   std_logic;
    pwm_write_data     : out   std_logic_vector(7 downto 0);
    pwm_read_data      : in    std_logic_vector(7 downto 0) );

end scanner;

architecture behavioral of scanner is

  -- scanning variables, fastest-changing first
  signal phase,  next_phase      : std_logic;
  signal strip,  next_strip      : std_logic_vector(2 downto 0);
  signal row,    next_row        : std_logic_vector(5 downto 0);
  signal col,    next_col        : std_logic_vector(2 downto 0);
  signal count,  next_count      : std_logic_vector(5 downto 0);

  -- internal versions of output signals
  signal odata,  next_odata      : std_logic_vector(7 downto 0);
  signal oreg,   next_oreg       : std_logic_vector(7 downto 0);
  signal oclk,   next_oclk       : std_logic;
  signal oload,  next_oload      : std_logic;
  signal ocolres, next_ocolres   : std_logic;
  signal ocolclk, next_ocolclk   : std_logic;
  signal pwm_bit                 : std_logic;

begin  -- behavioral

  SCAN_SYNC_PROCESS: process (clock_i, reset_i)
  begin
    if (reset_i = '1') then             -- load default values
      phase      <= '0';
      strip      <= (others =>'0');
      row        <= (others =>'0');
      col        <= (others =>'0');
      count      <= (others =>'0');
      odata      <= (others =>'0');
      oreg       <= (others =>'0');
      oclk       <= '0';
      oload      <= '0';
      ocolres    <= '0';
      ocolclk    <= '0';
    elsif (clock_i'event and clock_i = '1') then
      phase      <= next_phase;
      strip      <= next_strip;
      row        <= next_row;
      col        <= next_col;
      count      <= next_count;
      odata      <= next_odata;
      oreg       <= next_oreg;
      oclk       <= next_oclk;
      oload      <= next_oload;
      ocolres    <= next_ocolres;
      ocolclk    <= next_ocolclk;
    end if;
  end process SCAN_SYNC_PROCESS;

  SCAN_COMB_PROCESS:
  process (phase, strip, row, col, count, odata, oreg, oload, 
           strip_enable_i, pwm_read_data, scan_read_data, pwm_bit,
           disp_field, oclk, ocolres, ocolclk)

  begin

    -- clock is rising edge when strip goes from 3 to 4
    -- Actually we delay by half a cycle to provide setup and hold times
    -- since oreg data change when strip goes from 7 to 0
    if ((strip = "111" and phase = '1')
        or strip = "000" or strip = "001" or strip = "010"
        or (strip = "011" and phase = '0')) then
      next_oclk       <= '0';
    else
      next_oclk       <= '1';
    end if;

    -- Store the byte of output data for a while
    if (strip = "000" and phase = '0') then
      next_oreg       <= odata;
    else
      next_oreg       <= oreg;
    end if;

    -- create load pulse which is 4 clocks long, after last clock of col. 
    next_oload     <= '0';
    if ((row = "101111") and (strip(2) = '1')) then
      next_oload    <= '1';
    end if;

    -- create column counter reset pulse immediately after load pulse for col. 0
    -- column 0 is loaded at the start of column 1 time, but counter is wired to
    next_ocolclk         <= oload; 
    next_ocolres         <= '0';
    if (col="001") then
      next_ocolres       <= oload; 
    end if;

    -- count up variables to scan the TV screen; row counts down
    next_phase          <= not(phase);
    next_strip          <= strip;
    next_row            <= row;
    next_col            <= col;
    next_count          <= count;
    if (phase = '1') then
      next_strip        <= strip + "001";
      if (strip = "111") then
        next_row        <= row   - "000001";
        if (row = "000000") then
          next_row      <= "101111";
          next_col      <= col   + "001";
          if (col = "111") then
            next_count  <= count + "000001";
          end if;
        end if;
      end if;
    end if;

    -- build RAM address from above counter bits
    pwm_addr   <= row & strip & col;
    scan_addr   <= disp_field & row & strip & col;

    -- get pwm overflow bit for LED data
    pwm_bit     <= not(pwm_read_data(7));

    -- Phase 0 reads pixel RAM and pwm RAM
    if (phase = '0') then
      next_odata       <= odata;
      scan_enable      <= '1';
      pwm_enable       <= '1';
      pwm_write_en     <= '0';
      pwm_write_data   <= (others =>'0');
    -- Phase 1 computes new pwm RAM contents, writes to pwm RAM
    else
      -- inject this pixel's pwm data into the display data word
      next_odata       <= pwm_bit & odata(7 downto 1); 
      scan_enable      <= '0';
      pwm_enable       <= '1';
      pwm_write_en     <= '1';
      if (row = count) then
        pwm_write_data   <= "00" & scan_read_data(7 downto 2);   -- copy pixel to PWM value
--        pwm_write_data   <= '0' & row(5 downto 0) & '1';   -- TEST: data=row
      else
        pwm_write_data   <= pwm_read_data - "00000001";  -- decrement PWM value
      end if;
    end if;
  -- send outputs out to LED strips
  if (strip_enable_i = '1') then
    strip_data_o    <= oreg;
    strip_clk_o     <= oclk;
    strip_load_o    <= oload;
    strip_colres_o  <= ocolres;
    strip_colclk_o  <= ocolclk;
  else
    strip_data_o    <= (others=>'Z');
    strip_clk_o     <= 'Z';
    strip_load_o    <= 'Z';
    strip_colres_o  <= 'Z';
    strip_colclk_o  <= 'Z';
  end if;

  end process SCAN_COMB_PROCESS;


end behavioral;
